//----------------------------------------------------------------------------
// 
//	file: geninput.cc
// 
//	function: generate input 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Sat Sep 12 02:42:43 CDT 1998
// 
//	$Revision: 1.5 $ $Date: 2001/11/30 20:24:15 $
//	$Log: geninput.cc,v $
//	Revision 1.5  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.4  1999/02/09 07:42:25  yschoe
//	updated sine grating input
//
//	Revision 1.3  1999/01/05 09:20:50  yschoe
//	made gaussian() available as a static function.
//	wrote sine_grating routine -- must make it static too.
//
//	Revision 1.2  1998/11/17 12:42:30  yschoe
//	added ppm read routine.
//
//	Revision 1.1  1998/10/05 19:48:46  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------

#include "geninput.h"

//----------------------------------------------------------------------------
//	geninput() 
//	  - constructor : set input layer properties, and link to 
//	    actual array where input should be generated
//----------------------------------------------------------------------------
geninput::geninput(REAL **input, int cols, int rows, INP_TYPE type,
	INP_NORM norm, char *imgfilename, int imgcount) {

  // assume **input is already allocated somewhere else

  // input : is not allocated in this class. It uses a pre-allocated
  // 	matrix provided by the calling routine.
  geninput::input = input;

  geninput::cols = cols;
  geninput::rows = rows;
  geninput::inp_type  = type;
  geninput::norm_type = norm;

  // if type is image from file
  if (inp_type == INP_IMG) {
     geninput::imgcount = imgcount;
     // warning if imgcount == 0
     if (imgcount==0 || imgfilename == NULL) {
	warning("geninput",
	"image parameter error - no input will be generated");
     }
     // allocate memory for image
     images = new matrix<REAL>* [imgcount];
     for (int i=0; i<imgcount; ++i) 
	images[i] = new matrix<REAL>(cols,rows);
     // load images from file
     read_input(imgfilename);

  }

}

//----------------------------------------------------------------------------
//	~geninput() 
//	  - destructor : do nothing special
//----------------------------------------------------------------------------
geninput::~geninput() {

  if (inp_type == INP_IMG) {

     for (int i=0; i<imgcount; ++i) 
	delete images[i];

     delete [] images;

  }

}

//----------------------------------------------------------------------------
//	clear_input() 
//	  - clear input array with 0.0
//----------------------------------------------------------------------------
void geninput::clear_input() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	input[i][j] = 0.0;
     }
  }

}

//----------------------------------------------------------------------------
//	clear_rows() 
//	  - clear input array from rows "start" to "end"
//----------------------------------------------------------------------------
void geninput::clear_rows(int start, int end) {

  for (int i=0; i<cols; ++i) {
     for (int j=start; j<=end; ++j) {
	input[i][j] = 0.0;
     }
  }

}
//----------------------------------------------------------------------------
void geninput::clear_rows() {

  clear_rows((int)(rows/2),rows-1);

}

//----------------------------------------------------------------------------
//	add_input() 
//----------------------------------------------------------------------------
//	  - add input of type INP_SQR
void geninput::add_input(int cx, int cy, int radius) {

  int lowx = MAX(0,cx-radius);
  int lowy = MAX(0,cy-radius);
  int highx = MIN(cols-1, cx+radius);
  int highy = MIN(rows-1, cy+radius);

  REAL sum = 0.0;

  // set input
  for (int i=lowx; i<=highx; ++i) {
     for (int j=lowy; j<=highy; ++j) {
     	input[i][j] = 1.0;
	sum += norm_sum(1.0);
     }
  }

  // normalize
  normalize(lowx,highx,lowy,highy,sum);

}
//----------------------------------------------------------------------------
//	  - add input of type INP_ORI_BAR
void  geninput::add_input(int cx, int cy, int width, int height, REAL angle) {

  cout << "geninput::add_input(cx,cy,width,height,angle) not supported\n"; 

}

//----------------------------------------------------------------------------
//	  - add input of type INP_BLOB
void  geninput::add_input(int cx, int cy, REAL sigma) {


  // reuse oriented gaussian routine.
  add_input(cx,cy,sigma,sigma,0.0);

}

//----------------------------------------------------------------------------
//	  - add input of type INP_ORI_BLOB
void  geninput::add_input(int cx, int cy, REAL sigmax, REAL sigmay, REAL angle){

  // 'angle' is the orientation in degrees.
  // orientation 
  //           90
  //      180 _|_ 0
  //           |
  //          270        

  gaussian(input, cx, cy, sigmax, sigmay, angle);

}
//----------------------------------------------------------------------------
//	- add input of type INP_SIN
void  geninput::add_input(int cx, int cy, int freq, REAL phase, REAL angle) {

  sine_grating(input,cx,cy,freq,phase,angle);

}

//----------------------------------------------------------------------------
//	- add input of type INP_BLOBS @@@
void  geninput::add_input(int x, int y, REAL sigma, REAL angle) {

   // generate series of gaussian blobs of width sigma 
   // along a line with slope passing through (x,y)

   // first, generate a single small gaussian blob.

   // a rough size of the small matrix with the gaussian blob
   int radius = (int)(sigma*1.414213+0.5+2); // same as in the gaussian function

   matrix<REAL> *blob = new matrix<REAL>(2*radius+1,2*radius+1);
   blob->reset_matrix();
   gaussian(blob->mat, 2*radius+1,2*radius+1, radius, radius, sigma, sigma,0.0);

   // now stamp the blobs along the line with slope passing (x,y)

   matrix<REAL> *temp = new matrix<REAL>(cols,rows);
   temp->reset_matrix();

   // forward
   int cur_x = x, cur_y = y;
   int x_incr = (int)(radius*1.5*cos(angle*3.141592/180.0));
   int y_incr = (int)(radius*1.5*sin(angle*3.141592/180.0));
   while(cur_x<cols && cur_y<rows && cur_x>0 && cur_y>0) {
	temp->import_matrix(blob,cur_x,cur_y);
	cur_x += x_incr;
	cur_y += y_incr;
   }
   cur_x = x - x_incr;
   cur_y = y - y_incr;
   while(cur_x<cols && cur_y<rows && cur_x>0 && cur_y>0) {
	temp->import_matrix(blob,cur_x,cur_y);
	cur_x -= x_incr;
	cur_y -= y_incr;
   }
   temp->export_matrix(input);
   delete blob;
   delete temp;
   
}

//----------------------------------------------------------------------------
// 	- copy image to input matrix
void  geninput::add_input(int imgidx) {

  if (imgidx >= 0 && imgidx < imgcount ) {
  
     images[imgidx]->export_matrix(input);

  } else {

     warning("geninput::add_input","unknown image id");
  }
}  

//----------------------------------------------------------------------------
//	gaussian() 
//	  - generate gaussian spot
//----------------------------------------------------------------------------
void geninput::gaussian(REAL **input, int cols, int rows, int cx, int cy,
        REAL sigmax, REAL sigmay, REAL angle) {

  REAL radian = angle * M_PI / 180.0;
  register REAL ci, cj, xsq, ysq;

  int radius = (int)(sqrt(sigmax*sigmax+sigmay*sigmay)+0.5)+2;
  int lowx = MAX(0,cx-radius);
  int lowy = MAX(0,cy-radius);
  int highx = MIN(cols-1, cx+radius);
  int highy = MIN(rows-1, cy+radius);
  
  for (int i=lowx; i<=highx; ++i) {
     for (int j=lowy; j<=highy; ++j) {
        ci = cx-(REAL)i;
        cj = cy-(REAL)j;
        xsq = ci*cos(radian) + cj*sin(radian);
        ysq = -ci*sin(radian) + cj*cos(radian);
        xsq *= xsq;
        xsq /= (sigmax*sigmax);
        ysq *= ysq;
        ysq /= (sigmay*sigmay);

	REAL inp = exp(-(xsq+ysq));
        if (input[i][j]<inp) {
		input[i][j] = inp;
	}
     }
  }

}

//----------------------------------------------------------------------------
// for use within the class
void geninput::gaussian(REAL **input, int cx, int cy, 
	REAL sigmax, REAL sigmay, REAL angle) {

  gaussian(input,cols,rows,cx,cy,sigmax,sigmay,angle);

}

//----------------------------------------------------------------------------
//	normalize() 
//	  - normalize given area with given norm factor
//----------------------------------------------------------------------------
void geninput::normalize(int lowx, int highx, int lowy, int highy, REAL sum) {

  // normalize if necessary
  if (norm_type != INP_NORM_NONE) {

    REAL norm = 1.0/norm_factor(sum);

    for (int i=lowx; i<=highx; ++i) 
       for (int j=lowy; j<=highy; ++j) 
     	  input[i][j] *= norm;

  } 

}


//----------------------------------------------------------------------------
//	read_input() 
//	  - read input image(s) from file
//----------------------------------------------------------------------------
void geninput::read_input(char *filename) {

}

//----------------------------------------------------------------------------
//	singrating() 
//	  - generate sine grating with frequency, phase, and orientation, 
//	    centered at cx,cy
//	  - freq = 1 - ...
//	  - phase = 0-360 degrees
//	  - ori   = 0-360 degrees
//----------------------------------------------------------------------------
void geninput::sine_grating( REAL **input, int cols, int rows, int cx, int cy,
	int freq, REAL phase, REAL angle) {

  int 	x,y;
  REAL 	actual_angle, actual_freq, actual_phase;
  REAL 	sin_comp, cos_comp;

  actual_angle = M_PI*angle/180.0; // orientation
  actual_freq  = (REAL)freq/3.141592*1.5/2.0; // multiplier -- how many cycles?
  actual_phase = M_PI*phase*cols/360.0;

  // cout << angle << endl;

  // Minimum freq is Pi/2*cols, and maximum is PI/2RN*NFREQ 
  for(x=0; x<cols; x++)
    for(y=0 ; y<rows; y++) {
        cos_comp = (REAL)x * cos(actual_angle);
        sin_comp = (REAL)y * sin(actual_angle);
        input[x][y] = (sin((cos_comp + sin_comp)*actual_freq-actual_phase)
		       +1.0)/2.0;
    }

}
//----------------------------------------------------------------------------
// for use within the class
void geninput::sine_grating( REAL **input, int cx, int cy,
	int freq, REAL phase, REAL angle) {

  sine_grating(input,cols,rows,cx,cy,freq,phase,angle);

}
